@@ -1,6 +1,6 @@
 #----------------------------------------------------------------------
 # A minimal IMAP4 client for QRESYNC-capable servers
-# Copyright © 2015-2018 Guilhem Moulin <guilhem@fripost.org>
+# Copyright © 2015-2019 Guilhem Moulin <guilhem@fripost.org>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,18 +16,19 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #----------------------------------------------------------------------
 
-package Net::IMAP::InterIMAP v0.0.4;
+package Net::IMAP::InterIMAP v0.0.5;
+use v5.20.0;
 use warnings;
 use strict;
 
-use Compress::Raw::Zlib qw/Z_OK Z_FULL_FLUSH Z_SYNC_FLUSH MAX_WBITS/;
+use Compress::Raw::Zlib qw/Z_OK Z_STREAM_END Z_FULL_FLUSH Z_SYNC_FLUSH MAX_WBITS/;
 use Config::Tiny ();
 use Errno qw/EEXIST EINTR/;
-use Fcntl qw/F_GETFD F_SETFD FD_CLOEXEC/;
 use Net::SSLeay 1.73 ();
 use List::Util qw/all first/;
 use POSIX ':signal_h';
-use Socket qw/SOCK_STREAM IPPROTO_TCP AF_INET AF_INET6 SOCK_RAW :addrinfo/;
+use Socket qw/SOCK_STREAM SOCK_RAW SOCK_CLOEXEC IPPROTO_TCP SHUT_RDWR
+              AF_UNIX AF_INET AF_INET6 PF_UNSPEC :addrinfo/;
 
 use Exporter 'import';
 BEGIN {
@@ -35,14 +36,15 @@ BEGIN {
     Net::SSLeay::SSLeay_add_ssl_algorithms();
     Net::SSLeay::randomize();
 
-    our @EXPORT_OK = qw/xdg_basedir read_config compact_set $IMAP_text $IMAP_cond
+    our @EXPORT_OK = qw/xdg_basedir read_config compact_set
                         slurp is_dirty has_new_mails/;
 }
 
 
-# Regexes for RFC 3501's 'ATOM-CHAR', 'ASTRING-CHAR' and 'TEXT-CHAR'.
+# Regexes for RFC 3501's 'ATOM-CHAR', 'ASTRING-CHAR', 'list-char' and 'TEXT-CHAR'.
 my $RE_ATOM_CHAR    = qr/[\x21\x23\x24\x26\x27\x2B-\x5B\x5E-\x7A\x7C-\x7E]/;
 my $RE_ASTRING_CHAR = qr/[\x21\x23\x24\x26\x27\x2B-\x5B\x5D-\x7A\x7C-\x7E]/;
+my $RE_LIST_CHAR    = qr/[\x21\x23-\x27\x2A\x2B-\x5B\x5D-\x7A\x7C-\x7E]/;
 my $RE_TEXT_CHAR    = qr/[\x01-\x09\x0B\x0C\x0E-\x7F]/;
 
 my $RE_SSL_PROTO = qr/(?:SSLv[23]|TLSv1|TLSv1\.[0-3])/;
@@ -59,7 +61,7 @@ my %OPTIONS = (
     auth => qr/\A($RE_ATOM_CHAR+(?: $RE_ATOM_CHAR+)*)\z/,
     command => qr/\A(\P{Control}+)\z/,
     'null-stderr' => qr/\A(YES|NO)\z/i,
-    compress => qr/\A($RE_ATOM_CHAR+(?: $RE_ATOM_CHAR+)*)\z/,
+    compress => qr/\A(YES|NO)\z/i,
     SSL_protocols => qr/\A(!?$RE_SSL_PROTO(?: !?$RE_SSL_PROTO)*)\z/,
     SSL_fingerprint => qr/\A((?:[A-Za-z0-9]+\$)?\p{AHex}+)\z/,
     SSL_cipherlist => qr/\A(\P{Control}+)\z/,
@@ -91,11 +93,9 @@ sub xdg_basedir($$$$) {
     return $path if $path =~ /\A\//;
 
     my $basedir = $ENV{$xdg_variable};
-    unless (defined $basedir) {
-        my @getent = getpwuid($>);
-        $basedir = $getent[7] ."/". $default;
-    }
+    $basedir = ($ENV{HOME} // "") ."/". $default unless defined $basedir;
     die "No such directory: ", $basedir unless -d $basedir;
+
     $basedir .= "/".$subdir;
     $basedir =~ /\A(\/\p{Print}+)\z/ or die "Insecure $basedir";
     $basedir = $1;
@@ -203,6 +203,21 @@ sub compact_list(@) {
     return $set;
 }
 
+# with_set($set, $cmd)
+#   Split long commands over multiple subsets to avoid exceeding the server limit
+sub with_set($&) {
+    my ($set, $cmd) = @_;
+    my $max_length = 4096;
+    for (my $length = length($set); $length > $max_length;) {
+        my $l = rindex($set, ',', $max_length);
+        die unless $l > 0; # sanity check
+        $cmd->(substr($set, 0, $l));
+        $set = substr($set, ++$l);
+        $length -= $l;
+    }
+    return $cmd->($set);
+}
+
 
 # in_set($x, $set)
 #   Return true if the UID or sequence number $x belongs to the set $set.
@@ -215,7 +230,7 @@ sub in_set($$) {
             return 1 if $x == $1;
         }
         elsif ($r eq '*' or $r eq '*:*') {
-            warn "Assuming $x belongs to set $set!  (Dunno what \"*\" means.)";
+            warn "Assuming $x belongs to set $set! (Dunno what \"*\" means.)";
             return 1;
         }
         elsif ($r =~ /\A([0-9]+):\*\z/ or $r =~ /\A\*:([0-9]+)\z/) {
@@ -239,7 +254,7 @@ sub quote($) {
     if ($str =~ qr/\A$RE_ASTRING_CHAR+\z/) {
         return $str;
     }
-    elsif ($str =~ qr/\A$RE_TEXT_CHAR+\z/) {
+    elsif ($str =~ qr/\A$RE_TEXT_CHAR*\z/) {
         $str =~ s/([\x22\x5C])/\\$1/g;
         return "\"$str\"";
     }
@@ -281,7 +296,8 @@ our $IMAP_text;
 #
 #   - 'name': An optional instance name to include in log messages.
 #
-#   - 'logger-fd': An optional filehandle to use for debug output.
+#   - 'logger-fd': An optional filehandle to use for debug output
+#     (default: STDERR).
 #
 #   - 'keepalive': Whether to enable sending of keep-alive messages.
 #     (type=imap or type=imaps).
@@ -290,6 +306,7 @@ sub new($%) {
     my $class = shift;
     my $self = { @_ };
     bless $self, $class;
+    require 'Time/HiRes.pm' if defined $self->{'logger-fd'};
 
     # the IMAP state: one of 'UNAUTH', 'AUTH', 'SELECTED' or 'LOGOUT'
     # (cf RFC 3501 section 3)
@@ -303,18 +320,13 @@ sub new($%) {
 
     if ($self->{type} eq 'tunnel') {
         my $command = $self->{command} // $self->fail("Missing tunnel command");
-
-        pipe $self->{STDOUT}, my $wd or $self->panic("Can't pipe: $!");
-        pipe my $rd, $self->{STDIN}  or $self->panic("Can't pipe: $!");
-
-        my $pid = fork // $self->panic("Can't fork: $!");
+        socketpair($self->{S}, my $s, AF_UNIX, SOCK_STREAM|SOCK_CLOEXEC, PF_UNSPEC) or $self->panic("socketpair: $!");
+        my $pid = fork // $self->panic("fork: $!");
         unless ($pid) {
             # children
-            foreach (\*STDIN, \*STDOUT, $self->{STDIN}, $self->{STDOUT}) {
-                close $_ or $self->panic("Can't close: $!");
-            }
-            open STDIN,  '<&', $rd or $self->panic("Can't dup: $!");
-            open STDOUT, '>&', $wd or $self->panic("Can't dup: $!");
+            close($self->{S}) or $self->panic("Can't close: $!");
+            open STDIN,  '<&', $s or $self->panic("Can't dup: $!");
+            open STDOUT, '>&', $s or $self->panic("Can't dup: $!");
 
             my $stderr2;
             if ($self->{'null-stderr'} // 0) {
@@ -337,30 +349,24 @@ sub new($%) {
         }
 
         # parent
-        foreach ($rd, $wd) {
-            close $_ or $self->panic("Can't close: $!");
-        }
-        foreach (qw/STDIN STDOUT/) {
-            binmode($self->{$_}) // $self->panic("binmode: $!")
-        }
+        close($s) or $self->panic("Can't close: $!");
     }
     else {
         foreach (qw/host port/) {
             $self->fail("Missing option $_") unless defined $self->{$_};
         }
-        my $socket = defined $self->{proxy} ? $self->_proxify(@$self{qw/proxy host port/})
+        $self->{S} = defined $self->{proxy} ? $self->_proxify(@$self{qw/proxy host port/})
                                             : $self->_tcp_connect(@$self{qw/host port/});
         if (defined $self->{keepalive}) {
-            setsockopt($socket, Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
+            setsockopt($self->{S}, Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
                 or $self->fail("Can't setsockopt SO_KEEPALIVE: $!");
-            setsockopt($socket, Socket::IPPROTO_TCP, Socket::TCP_KEEPIDLE, 60)
+            setsockopt($self->{S}, Socket::IPPROTO_TCP, Socket::TCP_KEEPIDLE, 60)
                 or $self->fail("Can't setsockopt TCP_KEEPIDLE: $!");
         }
 
-        binmode($socket) // $self->panic("binmode: $!");
-        $self->_start_ssl($socket) if $self->{type} eq 'imaps';
-        $self->{$_} = $socket for qw/STDOUT STDIN/;
     }
+    binmode($self->{S}) // $self->panic("binmode: $!");
+    $self->_start_ssl($self->{S}) if $self->{type} eq 'imaps';
 
     # command counter
     $self->{_TAG} = 0;
@@ -390,11 +396,6 @@ sub new($%) {
     # are considered.
     $self->{_MODIFIED} = {};
 
-    if (defined $self->{'logger-fd'} and defined $self->{'logger-fd'}->fileno
-            and $self->{'logger-fd'}->fileno != fileno STDERR) {
-        require 'Time/HiRes.pm';
-    }
-
     # wait for the greeting
     my $x = $self->_getline();
     $x =~ s/\A\* (OK|PREAUTH) // or $self->panic($x);
@@ -412,15 +413,16 @@ sub new($%) {
         if ($self->{type} eq 'imap' and $self->{STARTTLS}) { # RFC 2595 section 5.1
             $self->fail("Server did not advertise STARTTLS capability.")
                 unless grep {$_ eq 'STARTTLS'} @caps;
-            $self->_start_ssl($self->{STDIN}) if $self->{type} eq 'imaps';
+            $self->_send('STARTTLS');
+            $self->_start_ssl($self->{S});
 
             # refresh the previous CAPABILITY list since the previous one could have been spoofed
             delete $self->{_CAPABILITIES};
             @caps = $self->capabilities();
         }
 
-        my @mechs = ('LOGIN', grep defined, map { /^AUTH=(.+)/i ? $1 : undef } @caps);
-        my $mech = (grep defined, map {my $m = $_; (grep {$m eq $_} @mechs) ? $m : undef}
+        my @mechs = ('LOGIN', grep defined, map { /^AUTH=(.+)/i ? uc($1) : undef } @caps);
+        my $mech = (grep defined, map {my $m = uc($_); (grep {$m eq $_} @mechs) ? $m : undef}
                                       split(/ /, $self->{auth}))[0];
         $self->fail("Failed to choose an authentication mechanism") unless defined $mech;
         $self->fail("Logins are disabled.") if ($mech eq 'LOGIN' or $mech eq 'PLAIN') and
@@ -448,8 +450,21 @@ sub new($%) {
             $self->fail("Unsupported authentication mechanism: $mech");
         }
 
+        my $dbg;
         delete $self->{password}; # no need to remember passwords
+        if (($self->{debug} // 0) == 1) {
+            $dbg = $self->{debug}--;
+            my $cmd = $command =~ /\A(LOGIN) / ? $1
+                    : $command =~ /\A(AUTHENTICATE \S+)(?: .*)?\z/ ? $1
+                    : $self->panic();
+            $self->logger('C: xxx ', $cmd, ' [REDACTED]');
+        }
         $self->_send($command, $callback);
+        if (defined $dbg) {
+            $self->logger('S: xxx ', $IMAP_text);
+            $self->{debug} = $dbg;
+        }
+
         unless ($IMAP_text =~ /\A\Q$IMAP_cond\E \[CAPABILITY /) {
             # refresh the CAPABILITY list since the previous one had only pre-login capabilities
             delete $self->{_CAPABILITIES};
@@ -517,51 +532,76 @@ sub stats($) {
 }
 
 
-# Log out when the Net::IMAP::InterIMAP object is destroyed.
+# Destroy a Net::IMAP::InterIMAP object.
 sub DESTROY($) {
+    local($., $@, $!, $^E, $?);
     my $self = shift;
     $self->{_STATE} = 'LOGOUT';
 
     Net::SSLeay::free($self->{_SSL}) if defined $self->{_SSL};
     Net::SSLeay::CTX_free($self->{_SSL_CTX}) if defined $self->{_SSL_CTX};
 
-    shutdown($self->{STDIN}, 2) if $self->{type} ne 'tunnel' and defined $self->{STDIN};
-    foreach (qw/STDIN STDOUT/) {
-        $self->{$_}->close() if defined $self->{$_} and $self->{$_}->opened();
+    if (defined (my $s = $self->{S})) {
+        # for type=tunnel we assume the child won't linger around once
+        # we close its standard input and output.
+        shutdown($s, SHUT_RDWR);
+        $s->close() if $s->opened();
     }
-
     $self->stats() unless $self->{quiet};
 }
 
 
 # $self->log($message, [...])
 # $self->logger($message, [...])
-#   Log a $message.  The latter method is used to log in the 'logger-fd', and
-#   add timestamps.
+#   Log a $message.  The latter method is used to log in the 'logger-fd'
+#   (and adds timestamps).
 sub log($@) {
     my $self = shift;
     return unless @_;
-    $self->logger(@_) if defined $self->{'logger-fd'} and defined $self->{'logger-fd'}->fileno
-        and $self->{'logger-fd'}->fileno != fileno STDERR;
-    my $prefix = $self->{name} // '';
-    $prefix .= "($self->{_SELECTED})" if $self->{_STATE} eq 'SELECTED';
-    $prefix .= ': ' unless $prefix eq '';
-    print STDERR $prefix, @_, "\n";
+    my $prefix = _logger_prefix($self);
+    if (defined (my $fd = $self->{'logger-fd'})) {
+        say $fd _date(), " ", $prefix, @_;
+    }
+    say STDERR $prefix, @_;
 }
 sub logger($@) {
     my $self = shift;
-    return unless @_ and defined $self->{'logger-fd'};
-    my $prefix = '';
-    if (defined $self->{'logger-fd'}->fileno and defined $self->{'logger-fd'}->fileno
-            and $self->{'logger-fd'}->fileno != fileno STDERR) {
-        my ($s, $us) = Time::HiRes::gettimeofday();
-        $prefix = POSIX::strftime("%b %e %H:%M:%S", localtime($s)).".$us";
-        $prefix .= ' ' if defined $self->{name} or $self->{_STATE} eq 'SELECTED';
+    return unless @_;
+    my $prefix = _logger_prefix($self);
+    if (defined (my $fd = $self->{'logger-fd'})) {
+        say $fd _date(), " ", $prefix, @_;
+    } else {
+        say STDERR $prefix, @_;
     }
-    $prefix .= $self->{name} if defined $self->{name};
-    $prefix .= "($self->{_SELECTED})" if $self->{_STATE} eq 'SELECTED';
-    $prefix .= ': ' unless $prefix eq '';
-    $self->{'logger-fd'}->say($prefix, @_);
+}
+sub _date() {
+    my ($s, $us) = Time::HiRes::gettimeofday();
+    my $t = POSIX::strftime("%b %e %H:%M:%S", localtime($s));
+    return "$t.$us"; # millisecond precision
+}
+
+# $self->_logger_prefix()
+#   Format a prefix for logging with printf(3)-like sequences:
+#     %n: the object name
+#     %m: mailbox, either explicit named or selected
+sub _logger_prefix($) {
+    my $self = shift;
+    my $format = $self->{'logger-prefix'} // return "";
+
+    my %seq = ( "%" => "%", m => $self->{mailbox}, n => $self->{name} );
+    $seq{m} //= $self->{_SELECTED} // die
+        if defined $self->{_STATE} and $self->{_STATE} eq 'SELECTED';
+
+    do {} while
+        # rewrite conditionals (loop because of nesting)
+        $format =~ s#%\? ([[:alpha:]]) \?
+                ( (?: (?> (?: [^%&?\\] | %[^?] | \\[&?\\] )+ ) | (?R) )* )
+         (?: \& ( (?: (?> (?: [^%&?\\] | %[^?] | \\[&?\\] )+ ) | (?R) )*) )?
+            \?# ($seq{$1} // "") ne "" ? $2 : ($3 // "") #agex;
+
+    $format =~ s#\\([&?\\])#$1#g; # unescape remaining '&', '?' and '\'
+    $format =~ s#%([%mn])# $seq{$1} #ge;
+    return $format;
 }
 
 
@@ -620,7 +660,7 @@ sub incapable($@) {
 
 
 # $self->search($criterion)
-#   Issue an UID SEARCH command with the given $criterion.  For the "normal"
+#   Issue a UID SEARCH command with the given $criterion.  For the "normal"
 #   UID SEARCH command from RFC 3501, return the list of matching UIDs;
 #   for the extended UID SEARCH command from RFC 4731 (ensuring ESEARCH
 #   capability is the caller's responsibility), return an optional "UID"
@@ -676,7 +716,7 @@ sub unselect($) {
 sub logout($) {
     my $self = shift;
     # don't bother if the connection is already closed
-    $self->_send('LOGOUT') if $self->{STDIN}->opened();
+    $self->_send('LOGOUT') if $self->{S}->opened();
     $self->{_STATE} = 'LOGOUT';
     undef $self;
 }
@@ -816,9 +856,10 @@ sub remove_message($@) {
     $self->fail("Server did not advertise UIDPLUS (RFC 4315) capability.")
         unless $self->_capable('UIDPLUS');
 
-    my $set = compact_set(@set);
-    $self->_send("UID STORE $set +FLAGS.SILENT (\\Deleted)");
-    $self->_send("UID EXPUNGE $set"); # RFC 4315 UIDPLUS
+    with_set(compact_set(@set), sub($) {
+        $self->_send("UID STORE $_[0] +FLAGS.SILENT (\\Deleted)");
+        $self->_send("UID EXPUNGE $_[0]"); # RFC 4315 UIDPLUS
+    });
 
     my %vanished = map {$_ => 1} @{$self->{_VANISHED}};
 
@@ -931,11 +972,13 @@ sub append($$@) {
 
 
 # $self->fetch($set, $flags, [$callback])
-#   Issue an UID FETCH command with the given UID $set, $flags, and
+#   Issue a UID FETCH command with the given UID $set, $flags, and
 #   optional $callback.
 sub fetch($$$;&) {
     my ($self, $set, $flags, $callback) = @_;
-    $self->_send("UID FETCH $set $flags", $callback);
+    return with_set($set, sub($) {
+        $self->_send("UID FETCH $_[0] $flags", $callback);
+    });
 }
 
 
@@ -967,7 +1010,7 @@ sub slurp($$$) {
     my $aborted = 0;
 
     my $rin = '';
-    vec($rin, fileno($_->{STDOUT}), 1) = 1 foreach @$selfs;
+    vec($rin, fileno($_->{S}), 1) = 1 foreach @$selfs;
 
     while (1) {
         # first, consider only unprocessed data without our own output
@@ -982,7 +1025,7 @@ sub slurp($$$) {
             next if $r == -1 and $! == EINTR; # select(2) was interrupted
             die "select: $!" if $r == -1;
             return $aborted if $r == 0; # nothing more to read (timeout reached)
-            @ready = grep {vec($rout, fileno($_->{STDOUT}), 1)} @$selfs;
+            @ready = grep {vec($rout, fileno($_->{S}), 1)} @$selfs;
             $timeout = $timeleft if $timeout > 0;
         }
 
@@ -1068,7 +1111,7 @@ sub set_cache($$%) {
         if ($k eq 'UIDVALIDITY') {
             # try to detect UIDVALIDITY changes early (before starting the sync)
             $self->fail("UIDVALIDITY changed! ($cache->{UIDVALIDITY} != $v)  ".
-                        "Need to invalidate the UID cache.")
+                        "Need to invalidate the UID cache for $mailbox.")
                 if defined $cache->{UIDVALIDITY} and $cache->{UIDVALIDITY} != $v;
         }
         $cache->{$k} = $v;
@@ -1172,16 +1215,15 @@ sub pull_updates($;$) {
     my $mailbox = $self->{_SELECTED} // $self->panic();
     my $pcache = $self->{_PCACHE}->{$mailbox};
 
-    my %modified;
     $self->_send("UID FETCH 1:".($pcache->{UIDNEXT}-1)." (MODSEQ FLAGS)")
         if $full and ($pcache->{UIDNEXT} // 1) > 1;
 
-    my @missing;
+    my %modified;
     while (%{$self->{_MODIFIED}}) {
+        my @missing;
         while (my ($uid,$v) = each %{$self->{_MODIFIED}}) {
-            # don't filter on the fly (during FETCH responses) because
-            # FLAG updates can arrive while processing pull_new_messages
-            # for instance
+            # don't filter on the fly (during FETCH responses) because FLAG updates
+            # can arrive while processing pull_new_messages() for instance
             if (defined $v->[1] and $v->[0] > 0) { # setting the MODSEQ to 0 forces a FETCH
                 next unless $uid              < ($pcache->{UIDNEXT} // 1)         # out of bounds
                         and ($full or $v->[0] > ($pcache->{HIGHESTMODSEQ} // 0)); # already seen
@@ -1191,8 +1233,11 @@ sub pull_updates($;$) {
             }
         }
         $self->{_MODIFIED} = {};
-        $self->_send("UID FETCH ".compact_set(@missing)." (MODSEQ FLAGS)") if @missing;
-        @missing = ();
+        # non-empty @missing indicates a discouraged (but allowed) CONDSTORE server behavior,
+        # cf. RFC 7162 sec. 3.1.3 ex. 8 and the comment in push_flag_updates() below
+        with_set(compact_set(@missing), sub($) {
+            $self->_send("UID FETCH $_[0] (MODSEQ FLAGS)")
+        }) if @missing;
     }
 
     # do that afterwards since the UID FETCH command above can produce VANISHED responses
@@ -1254,7 +1299,7 @@ sub pull_new_messages($$&@) {
         $range .= "$since:4294967295";
 
         $UIDNEXT = $cache->{UIDNEXT} // $self->panic(); # sanity check
-        $self->_send("UID FETCH $range ($attrs)", sub($) {
+        $self->fetch($range, "($attrs)", sub($) {
             my $mail = shift;
             $UIDNEXT = $mail->{UID} + 1 if $UIDNEXT <= $mail->{UID};
             $callback->($mail) if defined $callback;
@@ -1282,57 +1327,48 @@ sub push_flag_updates($$@) {
 
     my $mailbox = $self->{_SELECTED} // $self->panic();
     my $modseq = $self->{_PCACHE}->{$mailbox}->{HIGHESTMODSEQ} // $self->panic();
-    my $command = "UID STORE ".compact_set(@set)." FLAGS.SILENT ($flags) (UNCHANGEDSINCE $modseq)";
-
-    my %listed;
-    $self->_send($command, sub($){ $listed{shift->{UID}}++; });
 
     my %failed;
-    if ($IMAP_text =~ /\A\Q$IMAP_cond\E \[MODIFIED ([0-9,:]+)\] $RE_TEXT_CHAR+\z/) {
-        foreach (split /,/, $1) {
-            if (/\A([0-9]+)\z/) {
-                $failed{$1} = 1;
-            }
-            elsif (/\A([0-9]+):([0-9]+)\z/) {
-                my ($min, $max) = $1 < $2 ? ($1,$2) : ($2,$1);
-                $failed{$_} = 1 foreach ($min .. $max);
-            }
-            else {
-                $self->panic($_);
+    with_set(compact_set(@set), sub($) {
+        $self->_send("UID STORE $_[0] (UNCHANGEDSINCE $modseq) FLAGS.SILENT ($flags)");
+        if ($IMAP_text =~ /\A\Q$IMAP_cond\E \[MODIFIED ([0-9,:]+)\] $RE_TEXT_CHAR+\z/) {
+            foreach (split /,/, $1) {
+                if (/\A([0-9]+)\z/) {
+                    $failed{$1} = 1;
+                } elsif (/\A([0-9]+):([0-9]+)\z/) {
+                    my ($min, $max) = $1 < $2 ? ($1,$2) : ($2,$1);
+                    $failed{$_} = 1 foreach ($min .. $max);
+                } else {
+                    $self->panic($_);
+                }
             }
         }
-    }
+    });
 
     my @ok;
     foreach my $uid (@set) {
+        my $modified = $self->{_MODIFIED};
         if ($failed{$uid}) {
-            # $uid was listed in the MODIFIED response code
-            $self->{_MODIFIED}->{$uid} //= [ 0, undef ]; # will be downloaded again in pull_updates
-            delete $self->{_MODIFIED}->{$uid} if
-                # got a FLAG update for $uid; ignore it if it's $flags
-                defined $self->{_MODIFIED}->{$uid}->[1] and
-                $self->{_MODIFIED}->{$uid}->[1] eq $flags;
-        }
-        else {
-            # $uid wasn't listed in the MODIFIED response code
-            next unless defined $self->{_MODIFIED}->{$uid}; # already stored
-            $self->panic() unless defined $listed{$uid} and $listed{$uid} > 0; # sanity check
-            if ($listed{$uid} == 1) {
-                # ignore succesful update
-                delete $self->{_MODIFIED}->{$uid};
+            # $uid was listed in the MODIFIED response code from RFC 7162; will FETCH
+            # again in pull_updates(); per RFC 7162 sec. 3.1.3 $modified->{$uid} might not
+            # be defined ("nice" servers send an untagged FETCH response, cf. example 10,
+            # but they might omit it - allowed but discouraged CONDSTORE server behavior -
+            # cf. example 8)
+            $modified->{$uid} //= [ 0, undef ];
+        } elsif (defined (my $m = $modified->{$uid})) {
+            # received an untagged FETCH response, remove from the list of pending changes
+            # if the flag list was up to date (either implicitely or explicitely)
+            if (!defined $m->[1] or $m->[1] eq $flags) {
+                delete $modified->{$uid};
+                push @ok, $uid;
             }
-            elsif ($self->{_MODIFIED}->{$uid}->[1] and $self->{_MODIFIED}->{$uid}->[1] eq $flags) {
-                # got multiple FETCH responses for $uid, the last one with $flags
-                delete $self->{_MODIFIED}->{$uid};
-            }
-            push @ok, $uid;
         }
     }
 
     unless ($self->{quiet}) {
         $self->log("Updated flags ($flags) for UID ".compact_set(@ok)) if @ok;
         $self->log("Couldn't update flags ($flags) for UID ".compact_set(keys %failed).', '.
-                   "trying again later") if %failed;
+                   "will try again later") if %failed;
     }
     return keys %failed;
 }
@@ -1345,8 +1381,9 @@ sub push_flag_updates($$@) {
 sub silent_store($$$@) {
     my $self = shift;
     my $set = shift;
-    my $mod = shift;
-    $self->_send("UID STORE $set ${mod}FLAGS.SILENT (".join(' ', @_).")");
+    my $subcmd = shift . "FLAGS.SILENT";
+    my $flags = join(' ', @_);
+    with_set($set, sub($) { $self->_send("UID STORE $_[0] $subcmd ($flags)") });
 }
 
 
@@ -1359,7 +1396,7 @@ sub expunge($$) {
 
     $self->fail("Server did not advertise UIDPLUS (RFC 4315) capability.")
         unless $self->_capable('UIDPLUS');
-    $self->_send("UID EXPUNGE $set");
+    with_set($set, sub($) { $self->_send("UID EXPUNGE $_[0]") });
 }
 
 
@@ -1384,10 +1421,10 @@ sub _ssl_error($$@) {
 # RFC 3986 appendix A
 my $RE_IPv4 = do {
     my $dec = qr/[0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5]/;
-    qr/$dec(?:\.$dec){3}/o };
+    qr/$dec(?:\.$dec){3}/ };
 my $RE_IPv6 = do {
     my $h16  = qr/[0-9A-Fa-f]{1,4}/;
-    my $ls32 = qr/$h16:$h16|$RE_IPv4/o;
+    my $ls32 = qr/$h16:$h16|$RE_IPv4/;
     qr/                                  (?: $h16 : ){6} $ls32
       |                               :: (?: $h16 : ){5} $ls32
       | (?:                   $h16 )? :: (?: $h16 : ){4} $ls32
@@ -1397,7 +1434,7 @@ my $RE_IPv6 = do {
       | (?: (?: $h16 : ){0,4} $h16 )? ::                 $ls32
       | (?: (?: $h16 : ){0,5} $h16 )? ::                 $h16
       | (?: (?: $h16 : ){0,6} $h16 )? ::
-      /xo };
+      /x };
 
 
 # Opens a TCP socket to the given $host and $port.
@@ -1405,11 +1442,10 @@ sub _tcp_connect($$$) {
     my ($self, $host, $port) = @_;
 
     my %hints = (socktype => SOCK_STREAM, protocol => IPPROTO_TCP);
-    if ($host =~ qr/\A$RE_IPv4\z/o) {
+    if ($host =~ qr/\A$RE_IPv4\z/) {
         $hints{family} = AF_INET;
         $hints{flags} |= AI_NUMERICHOST;
-    }
-    elsif ($host =~ qr/\A\[($RE_IPv6)\]\z/o) {
+    } elsif ($host =~ qr/\A\[($RE_IPv6)\]\z/) {
         $host = $1;
         $hints{family} = AF_INET6;
         $hints{flags} |= AI_NUMERICHOST;
@@ -1420,7 +1456,7 @@ sub _tcp_connect($$$) {
 
     SOCKETS:
     foreach my $ai (@res) {
-        socket (my $s, $ai->{family}, $ai->{socktype}, $ai->{protocol}) or $self->panic("connect: $!");
+        socket (my $s, $ai->{family}, $ai->{socktype}|SOCK_CLOEXEC, $ai->{protocol}) or $self->panic("socket: $!");
 
         # timeout connect/read/write/... after 30s
         # XXX we need to pack the struct timeval manually: not portable!
@@ -1435,9 +1471,6 @@ sub _tcp_connect($$$) {
             next if $! == EINTR; # try again if connect(2) was interrupted by a signal
             next SOCKETS;
         }
-
-        my $flags = fcntl($s, F_GETFD, 0)       or $self->panic("fcntl F_GETFD: $!");
-        fcntl($s, F_SETFD, $flags | FD_CLOEXEC) or $self->panic("fcntl F_SETFD: $!");
         return $s;
     }
     $self->fail("Can't connect to $host:$port");
@@ -1590,7 +1623,7 @@ sub _ssl_verify($$$) {
 
             my $pkey = Net::SSLeay::X509_get_X509_PUBKEY($cert);
             unless (defined $pkey and Net::SSLeay::EVP_Digest($pkey, $type) eq $digest) {
-                $self->warn("Fingerprint doesn't match!  MiTM in action?");
+                $self->warn("Fingerprint doesn't match! MiTM in action?");
                 $ok = 0;
             }
         }
@@ -1690,6 +1723,7 @@ sub _start_ssl($$) {
     }
 
     @$self{qw/_SSL _SSL_CTX/} = ($ssl, $ctx);
+    undef $self; # the verify callback has reference to $self, free it now
 }
 
 
@@ -1703,7 +1737,7 @@ sub _getline($;$) {
     my $self = shift;
     my $len = shift // 0;
 
-    my ($stdout, $ssl) = @$self{qw/STDOUT _SSL/};
+    my ($stdout, $ssl) = @$self{qw/S _SSL/};
     $self->fail("Lost connection") unless $stdout->opened();
 
     my (@lit, @line);
@@ -1722,8 +1756,9 @@ sub _getline($;$) {
             $self->{_OUTRAWCOUNT} += $n;
 
             if (defined (my $i = $self->{_Z_INFLATE})) {
-                $i->inflate($buf, $self->{_OUTBUF}) == Z_OK or
-                    $self->panic("Inflation failed: ", $i->msg());
+                my $r = $i->inflate($buf, $self->{_OUTBUF});
+                $self->panic("Inflation failed: $r ", $i->msg())
+                    unless $r == Z_OK or $r == Z_STREAM_END;
             }
             else {
                 $self->{_OUTBUF} = $buf;
@@ -1784,7 +1819,7 @@ sub _update_cache_for($$%) {
         if ($k eq 'UIDVALIDITY') {
             # try to detect UIDVALIDITY changes early (before starting the sync)
             $self->fail("UIDVALIDITY changed! ($cache->{UIDVALIDITY} != $v)  ".
-                        "Need to invalidate the UID cache.")
+                        "Need to invalidate the UID cache for $mailbox.")
                 if defined $cache->{UIDVALIDITY} and $cache->{UIDVALIDITY} != $v;
             $self->{_PCACHE}->{$mailbox}->{UIDVALIDITY} //= $v;
         }
@@ -1840,8 +1875,8 @@ sub _cmd_extend($$) {
         $self->_cmd_extend_($args);
     }
     else {
-        # server supports LITERAL+: flush the command before each
-        # literal
+        # server doesn't supports LITERAL+: flush the command before
+        # each literal
         my ($offset, $litlen) = (0, 0);
         while ( (my $idx = index($$args, "\n", $offset+$litlen)) >= 0 ) {
             my $line = substr($$args, $offset, $idx+1-$offset);
@@ -1902,7 +1937,7 @@ sub _cmd_flush($;$$) {
     my $self = shift;
     $self->_cmd_extend_( $_[0] // \$CRLF );
     my $z_flush = $_[1] // Z_SYNC_FLUSH; # the flush point type to use
-    my ($stdin, $ssl) = @$self{qw/STDIN _SSL/};
+    my ($stdin, $ssl) = @$self{qw/S _SSL/};
 
     if ($self->{debug}) {
         # remove $CRLF and literals
@@ -2192,7 +2227,13 @@ sub _nstring($$) {
 # Parse and consume an RFC 3501 astring (1*ASTRING-CHAR / string).
 sub _astring($$) {
     my ($self, $stream) = @_;
-    return $$stream =~ s/\A($RE_ATOM_CHAR+)// ? $1 : $self->_string($stream);
+    return $$stream =~ s/\A$RE_ASTRING_CHAR+//p ? ${^MATCH} : $self->_string($stream);
+}
+
+# Parse and consume an RFC 3501 list-mailbox (1*list-char / string).
+sub _list_mailbox($$) {
+    my ($self, $stream) = @_;
+    return $$stream =~ s/\A$RE_LIST_CHAR+//p ? ${^MATCH} : $self->_string($stream);
 }
 
 # Parse and consume an RFC 3501 string (quoted / literal).
@@ -2327,7 +2368,7 @@ sub _resp($$;&$$) {
             # /!\ No bookkeeping since there is no internal cache mapping sequence numbers to UIDs
             if ($self->_enabled('QRESYNC')) {
                 $self->panic("$1 <= $cache->{EXISTS}") if $1 <= $cache->{EXISTS}; # sanity check
-                $self->fail("RFC 7162 violation!  Got an EXPUNGE response with QRESYNC enabled.");
+                $self->fail("RFC 7162 violation! Got an EXPUNGE response with QRESYNC enabled.");
             }
             # the new message was expunged before it was synced
             $self->{_NEW} = 0 if $self->{_NEW} == 1 and $cache->{EXISTS} == $1;
@@ -2364,11 +2405,11 @@ sub _resp($$;&$$) {
         elsif (s/\ALIST \((\\?$RE_ATOM_CHAR+(?: \\?$RE_ATOM_CHAR+)*)?\) ("(?:\\[\x22\x5C]|[\x01-\x09\x0B\x0C\x0E-\x21\x23-\x5B\x5D-\x7F])"|NIL) //) {
             my ($delim, $attrs) = ($2, $1);
             my @attrs = defined $attrs ? split(/ /, $attrs) : ();
-            my $mailbox = $self->_astring(\$_);
+            my $mailbox = $self->_list_mailbox(\$_);
             $self->panic($_) unless $_ eq '';
             $mailbox = 'INBOX' if uc $mailbox eq 'INBOX'; # INBOX is case-insensitive
             undef $delim if uc $delim eq 'NIL';
-            $delim =~ s/\A"(.*)"\z/$1/ if defined $delim;
+            $self->panic($_) if defined $delim and $delim !~ s/\A"\\?(.)"\z/$1/;
             $self->_update_cache_for($mailbox, DELIMITER => $delim);
             $self->_update_cache_for($mailbox, LIST_ATTRIBUTES => \@attrs);
             $callback->($mailbox, $delim, @attrs) if defined $callback and $cmd eq 'LIST';
@@ -2378,7 +2419,7 @@ sub _resp($$;&$$) {
             /\A \((\\?$RE_ATOM_CHAR+ [0-9]+(?: \\?$RE_ATOM_CHAR+ [0-9]+)*)?\)\z/ or $self->panic($_);
             my %status = split / /, $1;
             $mailbox = 'INBOX' if uc $mailbox eq 'INBOX'; # INBOX is case-insensitive
-            $self->panic("RFC 5465 violation!  Missing HIGHESTMODSEQ data item in STATUS response")
+            $self->panic("RFC 5465 violation! Missing HIGHESTMODSEQ data item in STATUS response")
                 if $self->_enabled('QRESYNC') and !defined $status{HIGHESTMODSEQ} and defined $cmd and
                    ($cmd eq 'NOTIFY' or $cmd eq 'slurp');
             $self->_update_cache_for($mailbox, %status);
